import os
import os.path as path
from typing import Union

from spirepy import Sample, Study
from spirepy.logger import logger


def download(item: Union[Study, Sample], target: str, output: str):
    """
    Dowload data from a SPIRE item.

    :param item: The item to be viewed (:class:`spirepy.sample.Sample` or :class:`spirepy.study.Study`).
    :type item: :class:`spirepy.sample.Sample` or :class:`spirepy.study.Study`

    :param target: What you want to view (metadata, antibiotic resistance annotations, manifest)
    :type target: str

    :param output: The output folder where the items will be downloaded.
    :type output: str
    """
    os.makedirs(output, exist_ok=True)
    if type(item) is Study:
        if target == "metadata":
            item.get_metadata().write_csv(path.join(output, f"{item.name}.csv"))
        elif target == "mags":
            item.download_mags(output)
        else:
            logger.error("No matching item for Study type")
    else:
        if target == "metadata":
            item.get_metadata().write_csv(path.join(output, f"{item.id}.csv"))
        elif target == "mags":
            item.download_mags(output)
        else:
            logger.error("No matching item")



import os
import os.path as path
import tarfile
import tempfile
import urllib

import polars as pl

from spirepy.data import genome_metadata


class Study:
    """
    A study from SPIRE.

    This class represents a study from the SPIRE database. It automatically
    fetches metadata and automates the initialization of samples to further use
    to obtain its genomic, geographical or other types of data provided by it.

    :param name: Internal ID for the study.
    :type name: str
    """

    def __init__(self, name: str):
        """Constructor method."""
        self.name = name
        self._metadata = None
        self._samples = None
        self._mags = None

    def get_metadata(self) -> pl.DataFrame:
        """Retrieve metadata for the study.

        :return: A Dataframe with the study's metadata.
        :rtype: :class:`polars.DataFrame`
        """
        if self._metadata is None:
            study_meta = pl.read_csv(
                f"https://spire.embl.de/spire/api/study/{self.name}?format=tsv",
                separator="\t",
            )
            self._metadata = study_meta
        return self._metadata

    def get_samples(self) -> list:
        """Retrive a list of samples for the study.

        :return: List of :class:`spirepy.sample.Sample` that belong to the study.
        :rtype: list
        """
        from spirepy.sample import Sample

        if self._samples is None:
            sample_list = []
            for s in self.get_metadata()["sample_id"].to_list():
                sample = Sample(s, self)
                sample_list.append(sample)
            self._samples = sample_list
        return self._samples

    def get_mags(self) -> pl.DataFrame:
        """Get a DataFrame with information regarding the MAGs.

        :return: A Dataframe with the study's MAGs.
        :rtype: :class:`polars.DataFrame`
        """
        if self._mags is None:
            genomes = genome_metadata()
            self._mags = genomes.filter(
                pl.col("derived_from_sample").is_in(
                    self.get_metadata()["sample_id"].to_list()
                )
            )
        return self._mags

    def download_assemblies(self, output: str):
        """Download the assemblies into a specified folder.

        :param output: Output folder to download the assemblies to.
        :type output: str
        """
        with tempfile.TemporaryDirectory() as tmpdir:
            tarfpath = path.join(tmpdir, f"{self.name}_assemblies.tar")
            urllib.request.urlretrieve(
                f"https://swifter.embl.de/~fullam/spire/compiled/{self.name}_spire_v1_assemblies.tar",
                tarfpath,
            )
            os.makedirs(output, exist_ok=True)
            with tarfile.open(tarfpath) as tar:
                tar.extractall(path.join(output, "assemblies"))

    def download_mags(self, output: str):
        """Download the MAGs into a specified folder.

        :param output: Output folder to download the MAGs to.
        :type output: str
        """
        with tempfile.TemporaryDirectory() as tmpdir:
            tarfpath = path.join(tmpdir, f"{self.name}_mags.tar")
            urllib.request.urlretrieve(
                f"https://swifter.embl.de/~fullam/spire/compiled/{self.name}_spire_v1_MAGs.tar",
                tarfpath,
            )
            os.makedirs(output, exist_ok=True)
            with tarfile.open(tarfpath) as tar:
                tar.extractall(path.join(output, "mags"))

    def download_genecalls(self, output: str):
        """Download the genecalls into a specified folder.

        :param output: Output folder to download the genecalls to.
        :type output: str
        """
        with tempfile.TemporaryDirectory() as tmpdir:
            tarfpath = path.join(tmpdir, f"{self.name}_genecalls.tar")
            urllib.request.urlretrieve(
                f"https://swifter.embl.de/~fullam/spire/genes_per_study/{self.name}_spire_v1_genecalls_fna.tar",
                tarfpath,
            )
            os.makedirs(output, exist_ok=True)
            with tarfile.open(tarfpath) as tar:
                tar.extractall(path.join(output, "genecalls"))

    def download_proteins(self, output: str):
        """Download the proteins into a specified folder.

        :param output: Output folder to download the proteins to.
        :type output: str
        """
        with tempfile.TemporaryDirectory() as tmpdir:
            tarfpath = path.join(tmpdir, f"{self.name}_genecalls.tar")
            urllib.request.urlretrieve(
                f"https://swifter.embl.de/~fullam/spire/genes_per_study/{self.name}_spire_v1_proteins_faa.tar",
                tarfpath,
            )
            os.makedirs(output, exist_ok=True)
            with tarfile.open(tarfpath) as tar:
                tar.extractall(path.join(output, "proteins"))



import re
import argparse

from spirepy import Study, Sample
from spirepy.cli import download, view


def maincall(input, action: str, target: str, output: str = None):
    if action == "view":
        view(input, target)
    else:
        download(input, target, output)


def main():
    parser = argparse.ArgumentParser(
        description="""
  _____ _____ _____ _____  ______             
 / ____|  __ \_   _|  __ \|  ____|            
| (___ | |__) || | | |__) | |__   _ __  _   _ 
 \___ \|  ___/ | | |  _  /|  __| | '_ \| | | |
 ____) | |    _| |_| | \ \| |____| |_) | |_| |
|_____/|_|   |_____|_|  \_\______| .__/ \__, |
                                 | |     __/ |
                                 |_|    |___/ 

Interact with the SPIRE[1] database.

[1] - Thomas S B Schmidt, Anthony Fullam, Pamela Ferretti, Askarbek Orakov,
Oleksandr M Maistrenko, Hans-Joachim Ruscheweyh, Ivica Letunic, Yiqian Duan,
Thea Van Rossum, Shinichi Sunagawa, Daniel R Mende, Robert D Finn, Michael Kuhn,
Luis Pedro Coelho, Peer Bork, SPIRE: a Searchable, Planetary-scale mIcrobiome
REsource, Nucleic Acids Research, Volume 52, Issue D1, 5 January 2024, Pages
D777–D783, https://doi.org/10.1093/nar/gkad943
""",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument(
        "--sample",
        dest="is_sample",
        action="store_true",
        help="The item you want to interact with is a sample",
    )
    parser.add_argument(
        "--study",
        dest="is_study",
        action="store_true",
        help="The item you want to interact with is a study",
    )
    subparsers = parser.add_subparsers(help="subcommand help", dest="action")
    # create the parser for the "view" command
    parser_view = subparsers.add_parser("view", help="view the data from an object")
    parser_view.add_argument(
        dest="target",
        choices=["metadata", "amr", "manifest", "eggnog", "mags"],
        action="store",
        help="target item to view",
    )
    parser_view.add_argument(
        "input", metavar="INPUT", nargs="+", help="Input (study or sample ID)", type=str
    )
    # create the parser for the "download" command
    parser_download = subparsers.add_parser(
        "download", help="download data from an item"
    )
    parser_download.add_argument(
        dest="target",
        choices=["mags", "proteins", "genecalls", "metadata"],
        action="store",
        help="target item to dowload",
    )
    parser_download.add_argument(
        "-o",
        "--output",
        dest="output",
        help="output folder; defaults to current folder",
        default="./",
    )
    parser_download.add_argument(
        "input", metavar="INPUT", nargs="+", help="Input (study or sample ID)", type=str
    )

    args = parser.parse_args()

    if args.is_sample:
        input = Sample(id=args.input[0])
        if args.action == "view":
            maincall(input, args.action, args.target)
        else:
            maincall(input, args.action, args.target, args.output)
    else:
        input = Study(name=args.input[0])
        if args.action == "view":
            maincall(input, args.action, args.target)
        else:
            maincall(input, args.action, args.target, args.output)


if __name__ == "__main__":
    main()



import unittest
from unittest.mock import patch

import polars as pl

from spirepy.data import cluster_metadata, genome_metadata


class TestDataFunctions(unittest.TestCase):
    @patch("polars.read_csv")
    def test_cluster_metadata_returns_polars_dataframe(self, mock_read_csv):
        mock_read_csv.return_value = pl.DataFrame(
            {"cluster_id": [1, 2], "description": ["Cluster 1", "Cluster 2"]}
        )

        result = cluster_metadata()

        self.assertIsInstance(result, pl.DataFrame)

    @patch("polars.read_csv")
    def test_genome_metadata_returns_polars_dataframe(self, mock_read_csv):
        mock_read_csv.return_value = pl.DataFrame(
            {"genome_id": [1, 2], "species": ["Species A", "Species B"]}
        )

        result = genome_metadata()

        self.assertIsInstance(result, pl.DataFrame)

    @patch("polars.read_csv")
    def test_cluster_metadata_caching(self, mock_read_csv):
        cluster_metadata.clear()
        mock_read_csv.return_value = pl.DataFrame({"a": [1]})

        cluster_metadata()
        cluster_metadata()

        mock_read_csv.assert_called_once()

    @patch("polars.read_csv")
    def test_genome_metadata_caching(self, mock_read_csv):
        genome_metadata.clear()
        mock_read_csv.return_value = pl.DataFrame({"a": [1]})

        genome_metadata()
        genome_metadata()

        mock_read_csv.assert_called_once()


if __name__ == "__main__":
    unittest.main()



from typing import Union

from spirepy import Sample, Study
from spirepy.logger import logger


def view(item: Union[Study, Sample], target: str):
    """
    View a SPIRE item.

    :param item: The item to be viewed (:class:`spirepy.sample.Sample` or :class:`spirepy.study.Study`).
    :type item: :class:`spirepy.sample.Sample` or :class:`spirepy.study.Study`

    :param target: What you want to view (metadata, antibiotic resistance annotations, manifest)
    :type target: str
    """

    if isinstance(item, Study):
        study_match = {
            "metadata": item.get_metadata,
            "mags": item.get_mags,
        }.get(target)

        if study_match:
            print(study_match())
        else:
            logger.error("No matching item for Study type")
    else: 
        sample_match = {
            "metadata": item.get_metadata,
            "mags": item.get_mags,
            "eggnog": item.get_eggnog_data,
            "amr": item.get_amr_annotations,
        }.get(target)

        if sample_match:
            print(sample_match())
        else:
            logger.error("No matching item")



import unittest
from unittest.mock import patch, MagicMock, call
import polars as pl
import pandas as pd
from polars.testing import assert_frame_equal

from spirepy import Sample, Study


class TestSample(unittest.TestCase):
    def setUp(self):
        """Set up a common Sample instance for all tests."""
        self.sample_id = "SAMPLE_01"
        self.study = Study(name="STUDY_A")
        self.sample = Sample(id=self.sample_id, study=self.study)

    def test_initialization_and_str(self):
        """Tests the __init__, __str__, and __repr__ methods."""
        self.assertEqual(self.sample.id, self.sample_id)
        self.assertEqual(self.sample.study, self.study)
        self.assertIsNone(self.sample._metadata)
        self.assertIsNone(self.sample._mags)
        self.assertIsNone(self.sample._eggnog_data)
        self.assertEqual(self.sample._amr_annotations, {})

        expected_str = f"Sample id: {self.sample_id} \tStudy: {self.study.name}"
        self.assertEqual(str(self.sample), expected_str)
        self.assertEqual(repr(self.sample), expected_str)

    @patch("spirepy.sample.pl.read_csv")
    def test_get_metadata(self, mock_read_csv: MagicMock):
        """Tests get_metadata for successful data retrieval and caching."""
        mock_data = pl.DataFrame({"sample_id": [self.sample_id]})
        mock_read_csv.return_value = mock_data

        # First call - should fetch data
        result1 = self.sample.get_metadata()
        expected_url = f"https://spire.embl.de/spire/api/sample/{self.sample_id}?format=tsv"
        mock_read_csv.assert_called_once_with(expected_url, separator="\t")
        assert_frame_equal(result1, mock_data)

        # Second call - should use cache
        result2 = self.sample.get_metadata()
        mock_read_csv.assert_called_once()  # Should still be called only once
        assert_frame_equal(result2, mock_data)

    @patch("spirepy.sample.genome_metadata")
    def test_get_mags(
        self,  mock_genome_metadata: MagicMock
    ):
        """Tests get_mags for data processing and caching."""
        # Mock input data
        mock_genome_data = pl.DataFrame(
            {
                "derived_from_sample": [self.sample_id, "OTHER_SAMPLE", self.sample_id],
                "spire_id": ["MAG_A", "MAG_B", "MAG_C"],
                "quality": [90, 95, 88],
            }
        )
        mock_genome_metadata.return_value = mock_genome_data

        # Expected result after filtering
        expected_mags = pl.DataFrame(
            {
                "derived_from_sample": [self.sample_id, self.sample_id],
                "spire_id": ["MAG_A", "MAG_C"],
                "quality": [90, 88],
            }
        )

        # First call - should process data
        result1 = self.sample.get_mags()
        mock_genome_metadata.assert_called_once()
        assert_frame_equal(result1, expected_mags)

        # Second call - should use cache
        result2 = self.sample.get_mags()
        mock_genome_metadata.assert_called_once()
        assert_frame_equal(result2, expected_mags)

    @patch("pandas.read_csv")
    def test_get_eggnog_data(self, mock_read_csv: MagicMock):
        """Tests get_eggnog_data for data retrieval and caching."""
        mock_df = pd.DataFrame({"gene": ["gene1"], "annotation": ["annot1"]})
        mock_read_csv.return_value = mock_df
        expected_data = pl.from_pandas(mock_df)

        # First call
        result1 = self.sample.get_eggnog_data()
        expected_url = f"https://spire.embl.de/download_eggnog/{self.sample_id}"
        mock_read_csv.assert_called_once_with(
            expected_url,
            sep="\t",
            skiprows=4,
            skipfooter=3,
            compression="gzip",
            engine="python",
        )
        assert_frame_equal(result1, expected_data)

        # Second call
        result2 = self.sample.get_eggnog_data()
        mock_read_csv.assert_called_once()
        assert_frame_equal(result2, expected_data)

    @patch("spirepy.sample.pl.read_csv")
    def test_get_amr_annotations(self, mock_read_csv: MagicMock):
        """Tests get_amr_annotations for all modes, caching, and invalid input."""
        mock_data = pl.DataFrame({"gene": ["gene_amr"], "resistance": ["drug_x"]})
        mock_read_csv.return_value = mock_data

        # Test 'deeparg' mode (default)
        result_deeparg = self.sample.get_amr_annotations(mode="deeparg")
        expected_url_deeparg = (
            f"https://spire.embl.de/download_deeparg/{self.sample_id}"
        )
        mock_read_csv.assert_called_with(expected_url_deeparg, separator="\t")
        assert_frame_equal(result_deeparg, mock_data)

        # Test caching for the same mode
        self.sample.get_amr_annotations(mode="deeparg")
        self.assertEqual(mock_read_csv.call_count, 1)

        # Test 'megares' mode - should be a new call
        result_megares = self.sample.get_amr_annotations(mode="megares")
        expected_url_megares = (
            f"https://spire.embl.de/download_abricate_megares/{self.sample_id}"
        )
        self.assertEqual(mock_read_csv.call_count, 2)
        mock_read_csv.assert_called_with(expected_url_megares, separator="\t")
        assert_frame_equal(result_megares, mock_data)

        # Test 'vfdb' mode - should be a new call
        result_vfdb = self.sample.get_amr_annotations(mode="vfdb")
        expected_url_vfdb = (
            f"https://spire.embl.de/download_abricate_vfdb/{self.sample_id}"
        )
        self.assertEqual(mock_read_csv.call_count, 3)
        mock_read_csv.assert_called_with(expected_url_vfdb, separator="\t")
        assert_frame_equal(result_vfdb, mock_data)

        # Verify all results are cached correctly
        self.assertEqual(len(self.sample._amr_annotations), 3)
        self.assertIn("deeparg", self.sample._amr_annotations)
        self.assertIn("megares", self.sample._amr_annotations)
        self.assertIn("vfdb", self.sample._amr_annotations)

    @patch("spirepy.sample.logger.error")
    def test_get_amr_annotations_invalid_mode(self, mock_logger_error: MagicMock):
        """Tests that an invalid mode returns None and logs an error."""
        result = self.sample.get_amr_annotations(mode="invalid_mode")
        self.assertIsNone(result)
        mock_logger_error.assert_called_once_with(
            "Invalid option, please choose one of the following: deeparg, megares, vfdb"
        )

    @patch("spirepy.sample.urllib.request.urlretrieve")
    @patch("spirepy.sample.os.makedirs")
    @patch.object(Sample, "get_mags")
    def test_download_mags(
        self,
        mock_get_mags: MagicMock,
        mock_makedirs: MagicMock,
        mock_urlretrieve: MagicMock,
    ):
        """Tests the download_mags functionality."""
        output_folder = "/fake/dir"
        mock_mags_data = pl.DataFrame({"spire_id": ["MAG_1", "MAG_2"]})
        mock_get_mags.return_value = mock_mags_data

        self.sample.download_mags(out_folder=output_folder)

        # Check that the output folder is created
        mock_makedirs.assert_called_once_with(output_folder, exist_ok=True)

        # Check that urlretrieve is called for each MAG
        expected_calls = [
            call(
                f"https://spire.embl.de/download_file/MAG_1",
                f"{output_folder}/MAG_1.fa.gz",
            ),
            call(
                f"https://spire.embl.de/download_file/MAG_2",
                f"{output_folder}/MAG_2.fa.gz",
            ),
        ]
        mock_urlretrieve.assert_has_calls(expected_calls, any_order=True)
        self.assertEqual(mock_urlretrieve.call_count, 2)

    @patch("spirepy.sample.urllib.request.urlretrieve")
    @patch("spirepy.sample.os.makedirs")
    @patch.object(Sample, "get_mags")
    def test_download_mags_no_mags(
        self,
        mock_get_mags: MagicMock,
        mock_makedirs: MagicMock,
        mock_urlretrieve: MagicMock,
    ):
        """Tests that no downloads are attempted if there are no MAGs."""
        output_folder = "/fake/dir"
        # Mock get_mags to return an empty DataFrame
        mock_get_mags.return_value = pl.DataFrame({"spire_id": []})

        self.sample.download_mags(out_folder=output_folder)

        # The directory should still be created
        mock_makedirs.assert_called_once_with(output_folder, exist_ok=True)
        # No download calls should be made
        mock_urlretrieve.assert_not_called()


if __name__ == "__main__":
    unittest.main(argv=["first-arg-is-ignored"], exit=False)



import logging
from rich.logging import RichHandler

formatter = logging.Formatter("%(message)s")
handler = RichHandler(level=logging.WARNING, markup=True)
handler.setFormatter(formatter)

logger = logging.getLogger("SPIREpy")
logger.addHandler(handler)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import sys
from pathlib import Path

# Add the project root to sys.path
docs_source = Path(__file__).parent
project_root = docs_source.parent.parent
sys.path.insert(0, str(project_root))

# Ensure autoapi output directory exists
autoapi_output = docs_source / 'autoapi'
autoapi_output.mkdir(exist_ok=True)

import spirepy


project = "SPIREpy"
copyright = "2025, Big Data Biology Lab"
author = "Big Data Biology Lab"
release = spirepy.__version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "nbsphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "autoapi.extension",
    "recommonmark"
]

autoapi_dirs = ['../../spirepy']
autoapi_type = 'python'
autoapi_root = 'autoapi'
autoapi_options = [
    'members',
    'undoc-members',
    'show-inheritance',
    'show-module-summary',
]
autoapi_ignore = ['*/__pycache__/*']
autoapi_keep_files = False

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
source_suffix = ['.rst', '.md', '.qmd']

# The main toctree document.
master_doc = "index"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ["**.ipynb_checkpoints"]

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_title = "version " + release
html_static_path = ["_static"]



import polars as pl
from joblib import Memory
from platformdirs import user_cache_dir

cache_dir = user_cache_dir("spirepy", "spirepy-dev")
memory = Memory(cache_dir, verbose=0)


@memory.cache
def cluster_metadata():
    """Fetches and caches the SPIRE cluster metadata from the remote server.

    This is slow for the first time, as it downloads a large file, but
    subsequent calls will use the cached version.

    :return: A DataFrame with the SPIRE cluster metadata.
    """
    return pl.read_csv(
        "https://swifter.embl.de/~fullam/spire/metadata/spire_v1_cluster_metadata.tsv.gz",
        separator="\t",
    )


@memory.cache
def genome_metadata():
    """Fetches and caches the SPIRE genome metadata from the remote server.

    This is slow for the first time, as it downloads a large file, but
    subsequent calls will use the cached version.

    :return: A DataFrame with the SPIRE genome metadata.
    """
    return pl.read_csv(
        "https://swifter.embl.de/~fullam/spire/metadata/spire_v1_genome_metadata.tsv.gz",
        separator="\t",
    )



from spirepy import Sample


class Genome:
    """
    A genome from SPIRE.
    """

    def __init__(self, id: str, sample: Sample):
        """
        Creates a new genome instance.
        """
        self.id = id
        self.sample = sample
        self._abundance = None



from typing import Union
import os
import os.path as path
import urllib.request

import polars as pl

from spirepy.data import genome_metadata
from spirepy.logger import logger
from spirepy.study import Study


class Sample:
    """
    A sample from SPIRE.

    This class represents a sample from the SPIRE database. It is designed to
    provide all the properties and methods to allow work with samples and
    provide tools for automation and scalability.

    :param id: Internal ID for the sample.
    :type id: str

    :param study: The :class:`spirepy.study.Study` to which the sample belongs to, defaults to :class:`None`.
    :type study: :class:`spirepy.study.Study`, optional
    """

    def __init__(self, id: str, study: Study = None):
        """Constructor method."""
        self.id = id
        self.study = study
        self._metadata = None
        self._mags = None
        self._eggnog_data = None
        self._amr_annotations = {}
        self._contig_depths = None

    def __str__(self):
        study_name = self.study.name if isinstance(self.study, Study) else None
        return f"Sample id: {self.id} 	Study: {study_name}"

    def __repr__(self):
        return self.__str__()

    def get_metadata(self) -> pl.DataFrame:
        """Retrieve the metadata for a sample.

        :return: A Dataframe with the sample's metadata.
        :rtype: :class:`polars.DataFrame`
        """
        if self._metadata is None:
            sample_meta = pl.read_csv(
                f"https://spire.embl.de/spire/api/sample/{self.id}?format=tsv",
                separator="\t",
            )
            self._metadata = sample_meta
        return self._metadata

    def get_mags(self) -> pl.DataFrame:
        """Retrieve the MAGs for a sample.

        :return: A Dataframe with the sample's MAGs.
        :rtype: :class:`polars.DataFrame`
        """
        if self._mags is None:
            genome_meta = genome_metadata()
            mags = genome_meta.filter(pl.col("derived_from_sample") == self.id)
            self._mags = mags
        return self._mags

    def get_eggnog_data(self) -> pl.DataFrame:
        """Retrive the EggNOG-mapper data for a sample.

        :return: A Dataframe with the sample's EggNOG-mapper data.
        :rtype: :class:`polars.DataFrame`
        """
        if self._eggnog_data is None:
            # We need to use pandas because polars does not support reading
            # files with a footer
            import pandas as pd

            egg = pd.read_csv(
                f"https://spire.embl.de/download_eggnog/{self.id}",
                sep="\t",
                skiprows=4,
                skipfooter=3,
                compression="gzip",
                engine="python",
            )
            eggnog_data = pl.from_pandas(egg)
            self._eggnog_data = eggnog_data
        return self._eggnog_data

    def get_amr_annotations(self, mode: str = "deeparg") -> Union[None, pl.DataFrame]:
        """Obtain the anti-microbial resistance annotations for the sample.

        :param mode: Tool to select the AMR data from. Options are deepARG (deeparg), abricate-megares (megares) and abricate-vfdb (vfdb); defaults to deepARG.
        :type mode: str, optional

        :return: A Dataframe with the sample's AMR data.
        :rtype: :class:`polars.DataFrame`
        """
        if mode not in self._amr_annotations:
            url = {
                "deeparg": f"https://spire.embl.de/download_deeparg/{self.id}",
                "megares": f"https://spire.embl.de/download_abricate_megares/{self.id}",
                "vfdb": f"https://spire.embl.de/download_abricate_vfdb/{self.id}",
            }.get(mode)
            if url is None:
                logger.error(
                    "Invalid option, please choose one of the following: deeparg, megares, vfdb"
                )
                return None
            amr = pl.read_csv(url, separator="\t")
            self._amr_annotations[mode] = amr
        return self._amr_annotations[mode]

    def get_contig_depths(self) -> pl.DataFrame:
        """Obtain the contig depth data for the sample.

        :return: A Dataframe with the sample's contig depth data.
        :rtype: :class:`polars.DataFrame`
        """
        if self._contig_depths is None:
            cont_depths = pl.read_csv(
                f"https://spire.embl.de/download_contig_depths/{self.id}",
                separator="\t",
            )
            self._contig_depths = cont_depths
        return self._contig_depths

    def download_mags(self, out_folder: str):
        """Download the MAGs into a specified folder.

        :param output: Output folder to download the MAGs to.
        :type output: str
        """
        os.makedirs(out_folder, exist_ok=True)
        for mag in self.get_mags()["spire_id"].to_list():
            urllib.request.urlretrieve(
                f"https://spire.embl.de/download_file/{mag}",
                path.join(out_folder, f"{mag}.fa.gz"),
            )



import unittest
from unittest.mock import patch, MagicMock
import polars as pl
from polars.testing import assert_frame_equal

from spirepy import Study, Sample


class TestStudy(unittest.TestCase):
    def setUp(self):
        """Set up a common Study instance for all tests."""
        self.study_name = "STUDY_XYZ"
        self.study = Study(name=self.study_name)

    def test_initialization(self):
        """Tests that the Study object is initialized correctly."""
        self.assertEqual(self.study.name, self.study_name)
        self.assertIsNone(self.study._metadata)
        self.assertIsNone(self.study._samples)
        self.assertIsNone(self.study._mags)

    @patch("spirepy.study.pl.read_csv")
    def test_get_metadata(self, mock_read_csv: MagicMock):
        """Tests metadata retrieval and caching."""
        mock_data = pl.DataFrame(
            {"study_id": [self.study_name], "sample_id": ["sample1"]}
        )
        mock_read_csv.return_value = mock_data

        # First call should fetch data
        result1 = self.study.get_metadata()
        expected_url = f"https://spire.embl.de/spire/api/study/{self.study_name}?format=tsv"
        mock_read_csv.assert_called_once_with(expected_url, separator="\t")
        assert_frame_equal(result1, mock_data)

        # Second call should use the cache
        result2 = self.study.get_metadata()
        mock_read_csv.assert_called_once()  # Should not be called again
        assert_frame_equal(result2, mock_data)

    @patch.object(Study, "get_metadata")
    def test_get_samples(self, mock_get_metadata: MagicMock):
        """Tests the creation of Sample objects and caching."""
        mock_metadata = pl.DataFrame({"sample_id": ["sample_A", "sample_B"]})
        mock_get_metadata.return_value = mock_metadata

        # First call should create Sample objects
        samples1 = self.study.get_samples()
        mock_get_metadata.assert_called_once()
        self.assertEqual(len(samples1), 2)
        self.assertIsInstance(samples1[0], Sample)
        self.assertEqual(samples1[0].id, "sample_A")
        self.assertEqual(samples1[0].study, self.study)  # Check back-reference
        self.assertEqual(samples1[1].id, "sample_B")

        # Second call should use the cache
        samples2 = self.study.get_samples()
        mock_get_metadata.assert_called_once()  # Should not be called again
        self.assertIs(samples1, samples2)  # Should be the exact same list object

    @patch("spirepy.study.genome_metadata")
    @patch.object(Study, "get_metadata")
    def test_get_mags(
        self, mock_get_metadata: MagicMock, mock_genome_metadata: MagicMock
    ):
        """Tests MAG filtering and caching."""
        # Mock the study's metadata
        mock_study_samples = ["sample_1", "sample_2"]
        mock_get_metadata.return_value = pl.DataFrame({"sample_id": mock_study_samples})

        # Mock the global genome metadata
        mock_all_genomes = pl.DataFrame(
            {
                "mag_id": ["MAG_A", "MAG_B", "MAG_C"],
                "derived_from_sample": ["sample_1", "sample_3", "sample_2"],
            }
        )
        mock_genome_metadata.return_value = mock_all_genomes

        # Define the expected filtered result
        expected_mags = pl.DataFrame(
            {
                "mag_id": ["MAG_A", "MAG_C"],
                "derived_from_sample": ["sample_1", "sample_2"],
            }
        )

        # First call should fetch and filter
        result1 = self.study.get_mags()
        mock_get_metadata.assert_called_once()
        mock_genome_metadata.assert_called_once()
        assert_frame_equal(result1, expected_mags)

        # Second call should use the cache
        result2 = self.study.get_mags()
        mock_get_metadata.assert_called_once()
        mock_genome_metadata.assert_called_once()
        assert_frame_equal(result2, expected_mags)

    @patch("spirepy.study.tarfile.open")
    @patch("spirepy.study.os.makedirs")
    @patch("spirepy.study.urllib.request.urlretrieve")
    @patch("spirepy.study.tempfile.TemporaryDirectory")
    def test_download_mags(
        self,
        mock_tempdir: MagicMock,
        mock_urlretrieve: MagicMock,
        mock_makedirs: MagicMock,
        mock_tar_open: MagicMock,
    ):
        """Tests the entire MAG download and extraction process."""
        output_dir = "/path/to/output"
        fake_temp_dir = "/tmp/fake_dir"

        # Configure the TemporaryDirectory mock
        mock_tempdir.return_value.__enter__.return_value = fake_temp_dir

        # Configure the tarfile.open mock
        mock_tar_object = MagicMock()
        mock_tar_open.return_value.__enter__.return_value = mock_tar_object

        # Execute the method
        self.study.download_mags(output=output_dir)

        # 1. Check that the download was called correctly
        expected_url = f"https://swifter.embl.de/~fullam/spire/compiled/{self.study_name}_spire_v1_MAGs.tar"
        expected_tar_path = f"{fake_temp_dir}/{self.study_name}_mags.tar"
        mock_urlretrieve.assert_called_once_with(expected_url, expected_tar_path)

        # 2. Check that the output directory was created
        mock_makedirs.assert_called_once_with(output_dir, exist_ok=True)

        # 3. Check that the tar file was opened
        mock_tar_open.assert_called_once_with(expected_tar_path)

        # 4. Check that the contents were extracted to the correct location
        expected_extract_path = f"{output_dir}/mags"
        mock_tar_object.extractall.assert_called_once_with(expected_extract_path)


if __name__ == "__main__":
    unittest.main(argv=["first-arg-is-ignored"], exit=False)



import unittest
from unittest.mock import patch
import polars as pl
from spirepy.data import cluster_metadata, genome_metadata, cache_dir
import shutil
import os

class TestDataCacheOnDisk(unittest.TestCase):
    def setUp(self):
        """Ensure the cache directory is clean before each test."""
        cluster_metadata.clear()
        genome_metadata.clear()
        if os.path.exists(cache_dir):
            shutil.rmtree(cache_dir)

    def tearDown(self):
        """Clean up the cache directory after each test."""
        if os.path.exists(cache_dir):
            shutil.rmtree(cache_dir)

    @patch('polars.read_csv')
    def test_disk_cache_creation(self, mock_read_csv):
        """Test that calling a cached function creates the on-disk cache."""
        mock_read_csv.return_value = pl.DataFrame({'a': [1]})
        
        self.assertFalse(os.path.exists(cache_dir))
        
        cluster_metadata()
        
        self.assertTrue(os.path.exists(cache_dir))
        self.assertTrue(len(os.listdir(cache_dir)) > 0)

if __name__ == '__main__':
    unittest.main()



import unittest
from unittest.mock import patch, MagicMock

from spirepy.cli.spire import main


class TestSpireCliMain(unittest.TestCase):
    @patch("spirepy.cli.spire.maincall")
    @patch("spirepy.cli.spire.argparse.ArgumentParser.parse_args")
    @patch("spirepy.cli.spire.Sample")
    def test_main_parses_sample_view_args(
        self, MockSample, mock_parse_args, mock_maincall
    ):
        """
        Tests that `main` correctly parses arguments for viewing a sample
        and calls maincall with a Sample object.
        """
        sample_id = "SAMPLE_ID_1"
        mock_args = MagicMock()
        mock_args.is_sample = True
        mock_args.is_study = False
        mock_args.action = "view"
        mock_args.target = "metadata"
        mock_args.input = [sample_id]
        mock_parse_args.return_value = mock_args

        mock_sample_instance = MagicMock()
        MockSample.return_value = mock_sample_instance

        main()

        MockSample.assert_called_once_with(id=sample_id)
        mock_maincall.assert_called_once_with(mock_sample_instance, "view", "metadata")

    @patch("spirepy.cli.spire.maincall")
    @patch("spirepy.cli.spire.argparse.ArgumentParser.parse_args")
    @patch("spirepy.cli.spire.Study")
    def test_main_parses_study_download_args(
        self, MockStudy, mock_parse_args, mock_maincall
    ):
        """
        Tests that `main` correctly parses arguments for downloading from a study
        and calls maincall with a Study object.
        """
        study_name = "STUDY_ID_X"
        output_dir = "/tmp/output"
        mock_args = MagicMock()
        mock_args.is_sample = False
        mock_args.is_study = True
        mock_args.action = "download"
        mock_args.target = "mags"
        mock_args.input = [study_name]
        mock_args.output = output_dir
        mock_parse_args.return_value = mock_args

        mock_study_instance = MagicMock()
        MockStudy.return_value = mock_study_instance

        main()

        MockStudy.assert_called_once_with(name=study_name)
        mock_maincall.assert_called_once_with(
            mock_study_instance, "download", "mags", output_dir
        )

    @patch("spirepy.cli.spire.maincall")
    @patch("spirepy.cli.spire.argparse.ArgumentParser.parse_args")
    @patch("spirepy.cli.spire.Study")
    def test_main_defaults_to_study(self, MockStudy, mock_parse_args, mock_maincall):
        """
        Tests that `main` defaults to creating a Study object if no type flag is given.
        """
        study_name = "DEFAULT_STUDY"
        mock_args = MagicMock()
        mock_args.is_sample = False
        mock_args.is_study = False
        mock_args.action = "view"
        mock_args.target = "mags"
        mock_args.input = [study_name]
        mock_parse_args.return_value = mock_args

        mock_study_instance = MagicMock()
        MockStudy.return_value = mock_study_instance

        main()

        MockStudy.assert_called_once_with(name=study_name)
        mock_maincall.assert_called_once_with(mock_study_instance, "view", "mags")


if __name__ == "__main__":
    unittest.main(argv=["first-arg-is-ignored"], exit=False)
